from collections import defaultdict
from collections.abc import Iterable
from typing import Any

from pdag._core import (
    CoreModel,
    FunctionRelationship,
    ParameterABC,
    SubModelRelationship,
)
from pdag._core.reference import ExecInfo
from pdag._utils import merge_two_set_dicts

from .model import (
    ConnectorABC,
    ExecInfoType,
    ExecutionModel,
    FunctionRelationshipInfo,
    ModelPathType,
    ParameterId,
    RelationshipId,
    StaticParameterId,
    StaticRelationshipId,
    TimeSeriesParameterId,
    TimeSeriesRelationshipId,
)
from .ref_resolver import resolve_ref


# TODO: Move this to a different module
def _iter_submodels_recursively(
    core_model: CoreModel,
    *,
    include_self: bool = False,
    _current_path: ModelPathType = (),
) -> Iterable[tuple[ModelPathType, CoreModel]]:
    assert core_model.is_hydrated(), "CoreModel must be hydrated."
    if include_self:
        yield _current_path, core_model
    for relationship in core_model.iter_all_relationships():
        if isinstance(relationship, SubModelRelationship):
            assert isinstance(relationship.name, str)
            yield from _iter_submodels_recursively(
                relationship.submodel,
                include_self=True,
                _current_path=(*_current_path, relationship.name),
            )


def _iter_function_relationships_recursively(
    core_model: CoreModel,
) -> Iterable[tuple[ModelPathType, CoreModel, FunctionRelationship[Any, Any]]]:
    """Iterate over all function relationships in the core model, including submodels.

    Yields tuples of the form (submodel_path, submodel, relationship).
    """
    assert core_model.is_hydrated(), "CoreModel must be hydrated."
    for submodel_path, submodel in _iter_submodels_recursively(
        core_model,
        include_self=True,
    ):
        for relationship in submodel.iter_all_relationships():
            if isinstance(relationship, FunctionRelationship):
                yield submodel_path, submodel, relationship


def _iter_submodel_relationships_recursively(
    core_model: CoreModel,
) -> Iterable[tuple[ModelPathType, CoreModel, SubModelRelationship]]:
    """Iterate over all submodel relationships in the core model, including submodels.

    Yields tuples of the form (submodel, relationship).
    """
    assert core_model.is_hydrated(), "CoreModel must be hydrated."
    for submodel_path, submodel in _iter_submodels_recursively(
        core_model,
        include_self=True,
    ):
        for relationship in submodel.iter_all_relationships():
            if isinstance(relationship, SubModelRelationship):
                yield submodel_path, submodel, relationship


def _iter_parameters_recursively(
    core_model: CoreModel,
) -> Iterable[tuple[ModelPathType, CoreModel, ParameterABC[Any]]]:
    """Iterate over all parameters in the core model, including submodels.

    Yields tuples of the form (submodel, parameter).
    """
    assert core_model.is_hydrated(), "CoreModel must be hydrated."
    for submodel_path, submodel in _iter_submodels_recursively(
        core_model,
        include_self=True,
    ):
        for parameter in submodel.iter_all_parameters():
            yield submodel_path, submodel, parameter


def _calculate_dependencies_of_time_series_function_relationship(
    relationship: FunctionRelationship[Any, Any],
    *,
    core_model: CoreModel,
    model_path: ModelPathType,
    n_time_steps: int,
) -> tuple[
    dict[ParameterId, set[RelationshipId]],  # input_parameter_id_to_relationship_ids
    dict[RelationshipId, set[ParameterId]],  # relationship_id_to_output_parameter_ids
    dict[
        RelationshipId,
        FunctionRelationshipInfo,
    ],  # relationship_id_to_function_relationship_info
]:
    assert relationship.at_each_time_step, "This function should only be called for time series relationships."
    assert isinstance(relationship.name, str)

    input_parameter_id_to_relationship_ids_dd: defaultdict[
        ParameterId,
        set[RelationshipId],
    ] = defaultdict(set)
    relationship_id_to_output_parameter_ids_dd: defaultdict[
        RelationshipId,
        set[ParameterId],
    ] = defaultdict(set)
    relationship_id_to_function_relationship_info: dict[
        RelationshipId,
        FunctionRelationshipInfo,
    ] = {}

    match relationship.includes_past, relationship.includes_future:
        case True, True:
            msg = "Relationships with both past and future dependencies are not supported."
            raise ValueError(msg)
        case True, False:
            time_steps = list(range(1, n_time_steps))
        case False, True:
            time_steps = list(range(n_time_steps - 1))
        case False, False:
            time_steps = list(range(n_time_steps))

    for time_step in time_steps:
        relationship_id = TimeSeriesRelationshipId(
            model_path=model_path,
            name=relationship.name,
            time_step=time_step,
        )
        input_args: dict[str, ConnectorABC | ExecInfoType] = {}
        for input_arg_name, input_parameter_ref in relationship.inputs.items():
            if isinstance(input_parameter_ref, ExecInfo):
                exec_info_type = ExecInfoType.from_exec_info(input_parameter_ref)
                input_args[input_arg_name] = exec_info_type
            else:
                input_connector = resolve_ref(
                    input_parameter_ref,
                    core_model=core_model,
                    model_path=model_path,
                    time_series_relationship=True,
                    time_step=time_step,
                )
                for input_parameter_id in input_connector.iter_parameter_ids():
                    input_parameter_id_to_relationship_ids_dd[input_parameter_id].add(
                        relationship_id,
                    )
                input_args[input_arg_name] = input_connector

        output_args: list[ConnectorABC] = []
        for output_parameter_ref in relationship.outputs:
            output_info = resolve_ref(
                output_parameter_ref,
                core_model=core_model,
                model_path=model_path,
                time_series_relationship=True,
                time_step=time_step,
            )
            relationship_id_to_output_parameter_ids_dd[relationship_id].update(
                output_info.iter_parameter_ids(),
            )
            output_args.append(output_info)

        relationship_id_to_function_relationship_info[relationship_id] = FunctionRelationshipInfo(
            function_relationship=relationship,
            input_parameter_info=input_args,
            output_parameter_info=tuple(output_args),
        )

    return (
        dict(input_parameter_id_to_relationship_ids_dd),
        dict(relationship_id_to_output_parameter_ids_dd),
        relationship_id_to_function_relationship_info,
    )


def _calculate_dependencies_of_static_function_relationship(
    relationship: FunctionRelationship[Any, Any],
    *,
    core_model: CoreModel,
    model_path: ModelPathType,
    n_time_steps: int,
) -> tuple[
    dict[ParameterId, set[RelationshipId]],  # input_parameter_id_to_relationship_ids
    dict[RelationshipId, set[ParameterId]],  # relationship_id_to_output_parameter_ids
    dict[
        RelationshipId,
        FunctionRelationshipInfo,
    ],  # relationship_id_to_function_relationship_info
]:
    assert not relationship.at_each_time_step, "This function should only be called for static relationships."
    assert isinstance(relationship.name, str)

    relationship_id = StaticRelationshipId(
        model_path=model_path,
        name=relationship.name,
    )
    input_parameter_ids: set[ParameterId] = set()
    output_parameter_ids: set[ParameterId] = set()

    input_args: dict[str, ConnectorABC | ExecInfoType] = {}
    for input_arg_name, input_parameter_ref in relationship.inputs.items():
        if isinstance(input_parameter_ref, ExecInfo):
            exec_info_type = ExecInfoType.from_exec_info(input_parameter_ref)
            input_args[input_arg_name] = exec_info_type
        else:
            input_info = resolve_ref(
                input_parameter_ref,
                core_model=core_model,
                model_path=model_path,
                time_series_relationship=False,
                n_time_steps=n_time_steps,
            )
            input_args[input_arg_name] = input_info
            input_parameter_ids.update(input_info.iter_parameter_ids())

    output_args: list[ConnectorABC] = []
    for output_parameter_ref in relationship.outputs:
        output_info = resolve_ref(
            output_parameter_ref,
            core_model=core_model,
            model_path=model_path,
            time_series_relationship=False,
            n_time_steps=n_time_steps,
        )
        output_parameter_ids.update(output_info.iter_parameter_ids())
        output_args.append(output_info)

    input_parameter_id_to_relationship_ids: dict[ParameterId, set[RelationshipId]] = {
        input_parameter_id: {relationship_id} for input_parameter_id in input_parameter_ids
    }
    relationship_id_to_output_parameter_ids: dict[RelationshipId, set[ParameterId]] = {
        relationship_id: output_parameter_ids,
    }
    relationship_id_to_function_relationship_info: dict[
        RelationshipId,
        FunctionRelationshipInfo,
    ] = {
        relationship_id: FunctionRelationshipInfo(
            function_relationship=relationship,
            input_parameter_info=input_args,
            output_parameter_info=tuple(output_args),
        ),
    }
    return (
        input_parameter_id_to_relationship_ids,
        relationship_id_to_output_parameter_ids,
        relationship_id_to_function_relationship_info,
    )


def _calculate_port_mapping_of_time_series_submodel_relationship(
    relationship: SubModelRelationship,
    *,
    core_model: CoreModel,
    model_path: ModelPathType,
    n_time_steps: int,
) -> dict[ParameterId, ParameterId]:
    # parent model input to sub-model input / sub-model output to parent model input
    port_mapping: dict[ParameterId, ParameterId] = {}

    assert isinstance(relationship.name, str)
    submodel_path = (*model_path, relationship.name)

    match relationship.includes_past, relationship.includes_future:
        case True, True:
            msg = "Relationships with both past and future dependencies are not supported."
            raise ValueError(msg)
        case True, False:
            time_steps = list(range(1, n_time_steps))
        case False, True:
            time_steps = list(range(n_time_steps - 1))
        case False, False:
            time_steps = list(range(n_time_steps))

    for time_step in time_steps:
        for (
            input_parameter_ref_inner,
            input_parameter_ref_outer,
        ) in relationship.inputs.items():
            input_parameter_inner = resolve_ref(
                input_parameter_ref_inner,
                core_model=relationship.submodel,
                model_path=submodel_path,
                time_series_relationship=True,
                time_step=time_step,
            )
            input_parameter_outer = resolve_ref(
                input_parameter_ref_outer,
                core_model=core_model,
                model_path=model_path,
                time_series_relationship=True,
                time_step=time_step,
            )
            for input_parameter_id_inner, input_parameter_id_outer in zip(
                input_parameter_inner.iter_parameter_ids(),
                input_parameter_outer.iter_parameter_ids(),
                strict=True,
            ):
                port_mapping[input_parameter_id_outer] = input_parameter_id_inner  # noqa: PERF403

        for (
            output_parameter_ref_inner,
            output_parameter_ref_outer,
        ) in relationship.outputs.items():
            output_parameter_inner = resolve_ref(
                output_parameter_ref_inner,
                core_model=relationship.submodel,
                model_path=submodel_path,
                time_series_relationship=True,
                time_step=time_step,
            )
            output_parameter_outer = resolve_ref(
                output_parameter_ref_outer,
                core_model=core_model,
                model_path=model_path,
                time_series_relationship=True,
                time_step=time_step,
            )
            for output_parameter_id_inner, output_parameter_id_outer in zip(
                output_parameter_inner.iter_parameter_ids(),
                output_parameter_outer.iter_parameter_ids(),
                strict=True,
            ):
                port_mapping[output_parameter_id_inner] = output_parameter_id_outer  # noqa: PERF403

    return port_mapping


def _calculate_port_mapping_of_static_submodel_relationship(
    relationship: SubModelRelationship,
    *,
    core_model: CoreModel,
    model_path: ModelPathType,
    n_time_steps: int,
) -> dict[ParameterId, ParameterId]:
    # parent model input to sub-model input / sub-model output to parent model input
    port_mapping: dict[ParameterId, ParameterId] = {}

    assert isinstance(relationship.name, str)
    submodel_path = (*model_path, relationship.name)

    for (
        input_parameter_ref_inner,
        input_parameter_ref_outer,
    ) in relationship.inputs.items():
        input_parameter_inner = resolve_ref(
            input_parameter_ref_inner,
            core_model=relationship.submodel,
            model_path=submodel_path,
            time_series_relationship=False,
            n_time_steps=n_time_steps,
        )
        input_parameter_outer = resolve_ref(
            input_parameter_ref_outer,
            core_model=core_model,
            model_path=model_path,
            time_series_relationship=False,
            n_time_steps=n_time_steps,
        )
        for input_parameter_id_inner, input_parmaeter_id_outer in zip(
            input_parameter_inner.iter_parameter_ids(),
            input_parameter_outer.iter_parameter_ids(),
            strict=True,
        ):
            port_mapping[input_parmaeter_id_outer] = input_parameter_id_inner  # noqa: PERF403

    for (
        output_parameter_ref_inner,
        output_parameter_ref_outer,
    ) in relationship.outputs.items():
        output_parameter_inner = resolve_ref(
            output_parameter_ref_inner,
            core_model=relationship.submodel,
            model_path=submodel_path,
            time_series_relationship=False,
            n_time_steps=n_time_steps,
        )
        output_parameter_outer = resolve_ref(
            output_parameter_ref_outer,
            core_model=core_model,
            model_path=model_path,
            time_series_relationship=False,
            n_time_steps=n_time_steps,
        )
        for output_parameter_id_inner, output_parameter_id_outer in zip(
            output_parameter_inner.iter_parameter_ids(),
            output_parameter_outer.iter_parameter_ids(),
            strict=True,
        ):
            port_mapping[output_parameter_id_inner] = output_parameter_id_outer  # noqa: PERF403

    return port_mapping


def create_exec_model_from_core_model(
    core_model: CoreModel,
    *,
    n_time_steps: int = 1,
) -> ExecutionModel:
    parameter_ids: set[ParameterId] = set()
    for model_path, _, parameter in _iter_parameters_recursively(core_model):
        assert isinstance(parameter.name, str)
        if parameter.is_time_series:
            parameter_ids.update(
                TimeSeriesParameterId(
                    model_path=model_path,
                    name=parameter.name,
                    time_step=time_step,
                )
                for time_step in range(n_time_steps)
            )
        else:
            parameter_ids.add(
                StaticParameterId(model_path=model_path, name=parameter.name),
            )

    relationships: dict[RelationshipId, FunctionRelationshipInfo] = {}
    input_parameter_id_to_relationship_ids: dict[ParameterId, set[RelationshipId]] = {}
    relationship_id_to_output_parameter_ids: dict[RelationshipId, set[ParameterId]] = {}
    port_mapping: dict[ParameterId, ParameterId] = {}

    for (
        model_path,
        model,
        function_relationship,
    ) in _iter_function_relationships_recursively(core_model):
        if function_relationship.at_each_time_step:
            (
                input_parameter_id_to_relationship_ids_local,
                relationship_id_to_output_parameter_ids_local,
                relationships_local,
            ) = _calculate_dependencies_of_time_series_function_relationship(
                function_relationship,
                core_model=model,
                model_path=model_path,
                n_time_steps=n_time_steps,
            )
        else:
            (
                input_parameter_id_to_relationship_ids_local,
                relationship_id_to_output_parameter_ids_local,
                relationships_local,
            ) = _calculate_dependencies_of_static_function_relationship(
                function_relationship,
                core_model=model,
                model_path=model_path,
                n_time_steps=n_time_steps,
            )
        relationships.update(relationships_local)

        input_parameter_id_to_relationship_ids = merge_two_set_dicts(
            input_parameter_id_to_relationship_ids,
            input_parameter_id_to_relationship_ids_local,
        )
        relationship_id_to_output_parameter_ids = merge_two_set_dicts(
            relationship_id_to_output_parameter_ids,
            relationship_id_to_output_parameter_ids_local,
        )

    for (
        model_path,
        model,
        sub_model_relationship,
    ) in _iter_submodel_relationships_recursively(core_model):
        if sub_model_relationship.at_each_time_step:
            port_mapping_local = _calculate_port_mapping_of_time_series_submodel_relationship(
                sub_model_relationship,
                core_model=model,
                model_path=model_path,
                n_time_steps=n_time_steps,
            )
        else:
            port_mapping_local = _calculate_port_mapping_of_static_submodel_relationship(
                sub_model_relationship,
                core_model=model,
                model_path=model_path,
                n_time_steps=n_time_steps,
            )
        port_mapping.update(port_mapping_local)

    return ExecutionModel(
        parameter_ids=parameter_ids,
        relationship_infos=relationships,
        input_parameter_id_to_relationship_ids=input_parameter_id_to_relationship_ids,
        relationship_id_to_output_parameter_ids=relationship_id_to_output_parameter_ids,
        port_mapping=port_mapping,
        n_time_steps=n_time_steps,
        _core_model=core_model,
    )



from typing import Annotated

import pdag


class UmbrellaModel(pdag.Model):
    """Evaluate which type of umbrella to take or not."""

    # Exogenous parameters
    rain_intensity = pdag.RealParameter(
        "rain_intensity",
        metadata={"XLRM": "X"},
        lower_bound=0.0,
        upper_bound=1.0,
    )

    # Levers
    policy = pdag.CategoricalParameter(
        "policy",
        ("take_umbrella", "take_travel_umbrella", "no_umbrella"),
        metadata={"XLRM": "L"},
    )

    # Performance metrics
    wetness = pdag.RealParameter(
        "wetness",
        metadata={"XLRM": "M"},
    )
    portability = pdag.RealParameter(
        "portability",
        metadata={"XLRM": "M"},
    )

    @pdag.relationship
    @staticmethod
    def calc_wetness(
        rain_intensity: Annotated[float, rain_intensity.ref()],
        policy: Annotated[str, policy.ref()],
    ) -> Annotated[float, wetness.ref()]:
        """Calculate the wetness based on the rain and policy."""
        match policy:
            case "take_umbrella":
                return rain_intensity * 0.1
            case "take_travel_umbrella":
                return rain_intensity * 0.2
            case "no_umbrella":
                return rain_intensity
            case _:
                msg = f"Unknown policy: {policy}"
                raise ValueError(msg)

    @pdag.relationship
    @staticmethod
    def calc_portability(
        policy: Annotated[str, policy.ref()],
    ) -> Annotated[float, portability.ref()]:
        """Calculate the portability based on the policy."""
        match policy:
            case "take_umbrella":
                return -1.0
            case "take_travel_umbrella":
                return -0.5
            case "no_umbrella":
                return 0.0
            case _:
                msg = f"Unknown policy: {policy}"
                raise ValueError(msg)


if __name__ == "__main__":
    from itertools import product

    import numpy as np
    from rich import print  # noqa: A004

    from pdag._experiment.multi_process import run_experiments
    from pdag._utils.random_string import generate_random_string

    core_model = UmbrellaModel.to_core_model()
    print("Core model:")
    print(core_model)

    exec_model = pdag.create_exec_model_from_core_model(core_model)
    print("Execution model:")
    print(exec_model)
    print("Input parameters:")
    print(exec_model.input_parameter_ids())

    # Run the model for one case
    results = pdag.execute_exec_model(
        exec_model,
        inputs={
            pdag.StaticParameterId((), "policy"): "take_umbrella",
            pdag.StaticParameterId((), "rain_intensity"): 0.5,
        },
    )
    print("Results:")
    print(results)

    # Run the model for multiple cases
    input_parameters = exec_model.input_parameters()
    scenario_parameters = {
        parameter_id: parameter
        for parameter_id, parameter in input_parameters.items()
        if parameter.metadata.get("XLRM", None) == "X"
    }
    decision_parameters = {
        parameter_id: parameter
        for parameter_id, parameter in input_parameters.items()
        if parameter.metadata.get("XLRM", None) == "L"
    }
    print("Scenario parameters:")
    print(scenario_parameters)
    print("Decision parameters:")
    print(decision_parameters)
    assert set(scenario_parameters) | set(decision_parameters) == set(input_parameters)

    n_scenarios = 100
    n_decisions = 100
    rng = np.random.default_rng(42)
    scenarios = pdag.distance_constrained_sampling(
        scenario_parameters,
        n_samples=n_scenarios,
        budget=1,
        rng=rng,
        uniform_in_distance=True,
    )
    decisions = pdag.sample_parameter_values(decision_parameters, n_samples=n_decisions, rng=rng)

    scenario_ids = [generate_random_string() for _ in range(n_scenarios)]
    decision_ids = [generate_random_string() for _ in range(n_decisions)]

    cases = (scenario | decision for scenario, decision in product(scenarios, decisions))
    metadata = (
        {"scenario_id": scenario_id, "decision_id": decision_id, "case_id": generate_random_string()}
        for scenario_id, decision_id in product(scenario_ids, decision_ids)
    )
    run_experiments(
        exec_model,
        cases,
        metadata=metadata,
        n_cases=n_scenarios * n_decisions,
        parquet_file_path="umbrella.parquet",
    )



from collections.abc import Hashable, Iterable
from dataclasses import dataclass, field
from typing import Annotated, Any, cast

from typing_extensions import Doc

from .collection import CollectionABC
from .parameter import ParameterABC
from .reference import CollectionRef, ParameterRef, ReferenceABC
from .relationship import RelationshipABC


@dataclass
class CoreModel:
    name: Annotated[str, Doc("Name of the model.")]
    parameters: Annotated[
        dict[str, ParameterABC[Any]],
        Doc(
            "Mapping of parameter names to parameters. Parameters contained in collections are not included.",
        ),
    ]
    relationships: Annotated[
        dict[str, RelationshipABC],
        Doc(
            "Mapping of relationship names to relationships. Relationships contained in collections are not included.",
        ),
    ]
    collections: Annotated[
        dict[str, CollectionABC[Hashable, ParameterABC[Any] | RelationshipABC]],
        Doc("Mapping of collection names to collections."),
    ]

    _parameter_dict: dict[str, ParameterABC[Any]] = field(init=False)
    _relationship_dict: dict[str, RelationshipABC] = field(init=False)

    def __post_init__(self) -> None:
        self._parameter_dict = {param.name: param for param in self.iter_all_parameters()}
        self._relationship_dict = {relationship.name: relationship for relationship in self.iter_all_relationships()}

    def is_hydrated(self) -> bool:
        """Check if all parameters, collections, and relationships are hydrated."""
        return (
            all(parameter.is_hydrated() for parameter in self.parameters.values())
            and all(collection.is_hydrated() for collection in self.collections.values())
            and all(relationship.is_hydrated() for relationship in self.relationships.values())
        )

    def is_dynamic(self) -> bool:
        """Check if any parameter is dynamic."""
        return any(parameter.is_time_series for parameter in self.parameters.values())

    def iter_all_parameters(self) -> Iterable[ParameterABC[Any]]:
        """Yield all parameters contained in the model, including those in collections.

        It does NOT include parameters in sub-models.
        """
        yield from self.parameters.values()
        for collection in self.collections.values():
            if collection.item_type == "parameter":
                yield from cast("CollectionABC[Hashable, ParameterABC[Any]]", collection).values()

    def iter_all_relationships(self) -> Iterable[RelationshipABC]:
        """Yield all relationships contained in the model, including those in collections.

        It does NOT include relationships in sub-models.
        """
        yield from self.relationships.values()
        for collection in self.collections.values():
            if collection.item_type == "relationship":
                yield from cast("CollectionABC[Hashable, RelationshipABC]", collection).values()

    def get_relationship(self, name: str) -> RelationshipABC:
        return self._relationship_dict[name]

    def get_parameter(self, name: str) -> ParameterABC[Any]:
        return self._parameter_dict[name]

    def get_object_from_ref(
        self,
        ref: ReferenceABC,
    ) -> ParameterABC[Any] | CollectionABC[Any, Any]:
        if isinstance(ref, ParameterRef):
            return self.get_parameter_from_ref(ref)
        if isinstance(ref, CollectionRef):
            return self.get_collection_from_ref(ref)
        msg = f"Reference type {type(ref)} is not supported."
        raise TypeError(msg)

    def get_parameter_from_ref(self, ref: ParameterRef) -> ParameterABC[Any]:
        return self.parameters[ref.name]

    def get_collection_from_ref(self, ref: CollectionRef[Any]) -> CollectionABC[Any, Any]:
        return self.collections[ref.name]


@dataclass
class Module:
    pre_models: str
    models: list[CoreModel]  # Assuming topological order
    post_models: str

    def is_hydrated(self) -> bool:
        return all(model.is_hydrated() for model in self.models)



import importlib
import importlib.metadata
import logging
from itertools import chain
from pathlib import Path
from typing import TYPE_CHECKING, Annotated, NoReturn

import typer
from rich.console import Console
from watchfiles import watch as _watch

from pdag._exec.to_exec_model import create_exec_model_from_core_model
from pdag._export.dot import export_dot

if TYPE_CHECKING:
    from pdag._notation import Model

logger = logging.getLogger(__name__)

console = Console()
err_console = Console(stderr=True)

app = typer.Typer()


def version_callback(value: bool) -> None:  # noqa: FBT001
    if value:
        version = importlib.metadata.version("pdag")
        typer.echo(f"pdag {version}")
        raise typer.Exit


@app.callback()
def main(
    version: Annotated[
        bool | None,
        typer.Option("--version", help="Show the version and exit", is_eager=True, callback=version_callback),
    ] = None,
) -> None:
    """Pdag CLI."""


@app.command()
def watch(
    model: Annotated[str, typer.Argument(..., help="Model specified as 'module_name:ModelName'")],
    output_path: Annotated[Path, typer.Argument(..., help="Path to output file")],
    *,
    show_locals_on_error: Annotated[bool, typer.Option("--show-locals", help="Show locals on error")] = False,
    try_conversion_to_exec_model: Annotated[
        bool,
        typer.Option("--try-conversion", help="Try conversion to exec model"),
    ] = True,
    n_time_steps: Annotated[int, typer.Option("--n-time-steps", help="Number of time steps for exec model")] = 1,
) -> NoReturn:
    """Watch a model."""
    module_str, _, attr_str = model.partition(":")
    if not attr_str:
        msg = "Model must be specified as 'module_name:ModelName'"
        raise typer.BadParameter(msg)

    try:
        module = importlib.import_module(module_str)
    except Exception as e:
        msg = f"Error importing module {module_str}: {e}"
        raise ValueError(msg) from e

    if module.__file__ is None:
        msg = f"Module {module_str} has no __file__ attribute"
        raise ValueError(msg)

    # Chain a 1-element iterable to execute immediately before the first change is detected
    for _ in chain((None,), _watch(module.__file__)):
        console.clear()
        err_console.clear()

        err_console.print(f":eyes: Watching changes in {module.__file__} for {module_str}:{attr_str}...")

        with err_console.status(f"Loading {module_str}:{attr_str}..."):
            try:
                module = importlib.reload(module)
                pdag_model: Model = getattr(module, attr_str)
            except Exception:  # noqa: BLE001
                err_console.print_exception(show_locals=show_locals_on_error)
                continue

            try:
                core_model = pdag_model.to_core_model()
            except Exception:  # noqa: BLE001
                err_console.print_exception(show_locals=show_locals_on_error)
                continue

            try:
                export_dot(core_model, output_path)
            except Exception:  # noqa: BLE001
                err_console.print_exception(show_locals=show_locals_on_error)
                continue

        err_console.print(f":white_check_mark: Exported {core_model.name} to {output_path}")

        if try_conversion_to_exec_model:
            with err_console.status(f"Creating exec model from {core_model.name} with n_time_steps={n_time_steps}..."):
                try:
                    create_exec_model_from_core_model(
                        core_model,
                        n_time_steps=n_time_steps,
                    )
                except Exception:  # noqa: BLE001
                    err_console.print_exception(show_locals=show_locals_on_error)
                    continue
            err_console.print(f":white_check_mark: Created exec model from {core_model.name}")
            continue

    msg = "`pdag watch` should never return"
    raise RuntimeError(msg)



import re
from collections.abc import Hashable
from pathlib import Path
from typing import Any, Literal

import pydot

from pdag._core import CollectionABC, CoreModel, ExecInfo
from pdag._core.collection import key_to_str
from pdag._core.parameter import ParameterABC
from pdag._core.relationship import RelationshipABC
from pdag._utils.random_string import generate_random_string


def sanitize_for_html_id(part: str) -> str:
    # Replace non-alphanumeric with dash
    part = re.sub(r"[^a-zA-Z0-9_-]", "-", part)
    # Remove leading/trailing dashes
    part = part.strip("-")
    # Ensure starts with a letter or underscore
    if not re.match(r"^[a-zA-Z_]", part):
        part = "_" + part
    return part


def tuple_to_html_id(parts: tuple[str, ...]) -> str:
    sanitized = [sanitize_for_html_id(p) for p in parts]
    return "__".join(sanitized)


def _mapping_node(
    collection: CollectionABC[Hashable, ParameterABC[Any] | RelationshipABC],
) -> pydot.Node:
    peripheries = 1
    if collection.item_type == "parameter" and collection.is_time_series():
        peripheries = 2
    if collection.item_type == "relationship":
        first_relationship = next(iter(collection.values()))
        assert isinstance(first_relationship, RelationshipABC)
        if first_relationship.at_each_time_step:
            peripheries = 2

    header_row = "<TR><TD BGCOLOR='lightgrey'>"
    header_row += f"<B>{collection.name}</B></TD></TR>"

    def _item_to_row(key: Hashable) -> str:
        key_label = key_to_str(key, make_one_element_tuple_scalar=True)
        return f"<TR><TD PORT='{key_label}'>{key_label}</TD></TR>"

    rows = header_row + "".join(_item_to_row(key) for key in collection.keys())  # noqa: SIM118
    table = f'<TABLE BORDER="0" CELLBORDER="1" CELLSPACING="0">{rows}</TABLE>'
    label = f"<{table}>"

    return pydot.Node(
        collection.name,
        shape="oval" if collection.item_type == "parameter" else "box",
        peripheries=peripheries,
        label=label,
    )


def to_dot_graph(core_model: CoreModel, *, model_element_id: str | None = None) -> pydot.Dot:  # noqa: C901, PLR0912
    model_element_id = core_model.name if model_element_id is None else model_element_id

    graph = pydot.Dot(core_model.name, graph_type="digraph", id=model_element_id)
    added_exec_infos: set[str] = set()

    for name, parameter in core_model.parameters.items():
        element_id = tuple_to_html_id((model_element_id, name))
        if parameter.is_time_series:
            graph.add_node(pydot.Node(name, shape="oval", peripheries=2, id=element_id))
        else:
            graph.add_node(pydot.Node(name, shape="oval", id=element_id))

    for name, relationship in core_model.relationships.items():
        element_id = tuple_to_html_id((model_element_id, name))
        if relationship.at_each_time_step:
            graph.add_node(pydot.Node(name, shape="box", peripheries=2, id=element_id))
        else:
            graph.add_node(pydot.Node(name, shape="box", id=element_id))

        for input_ref in relationship.iter_input_refs():
            if isinstance(input_ref, ExecInfo):
                element_id = tuple_to_html_id((model_element_id, name, input_ref.attribute, generate_random_string()))
                if input_ref.attribute not in added_exec_infos:
                    graph.add_node(pydot.Node(input_ref.attribute, shape="diamond", id=element_id))
                    added_exec_infos.add(input_ref.attribute)
                graph.add_edge(pydot.Edge(input_ref.attribute, name))
            else:
                style = "dashed" if input_ref.previous else "solid"
                graph.add_edge(pydot.Edge(input_ref.name, name, style=style))

        for output_ref in relationship.iter_output_refs():
            style = "dashed" if output_ref.next else "solid"
            graph.add_edge(pydot.Edge(name, output_ref.name, style=style))

    for name, collection in core_model.collections.items():
        graph.add_node(_mapping_node(collection))

        # The elements of the set are tuples of the form: (execinfo_or_param, name, is_delayed)
        input_params: set[tuple[Literal["execinfo", "param"], str, bool]] = set()
        output_params: set[tuple[Literal["execinfo", "param"], str, bool]] = set()

        for item in collection.values():
            if isinstance(item, RelationshipABC):
                for input_ref in item.iter_input_refs():
                    if isinstance(input_ref, ExecInfo):
                        input_params.add(("execinfo", input_ref.attribute, False))
                    else:
                        input_params.add(("param", input_ref.name, input_ref.previous))

                for output_ref in item.iter_output_refs():
                    output_params.add(("param", output_ref.name, output_ref.next))

        for i in input_params:
            if i[0] == "execinfo":
                attr = i[1]
                if attr not in added_exec_infos:
                    graph.add_node(pydot.Node(attr, shape="diamond"))
                    added_exec_infos.add(attr)
                graph.add_edge(pydot.Edge(attr, name))
            elif i[0] == "param":
                param = i[1]
                style = "dashed" if i[2] else "solid"
                graph.add_edge(pydot.Edge(param, name, style=style))
            else:
                msg = f"Unknown type: {i[0]}"
                raise ValueError(msg)

        for o in output_params:
            param = o[1]
            style = "dashed" if o[2] else "solid"
            graph.add_edge(pydot.Edge(name, param, style=style))

    return graph


def export_dot(core_model: CoreModel, path: Path) -> None:
    graph = to_dot_graph(core_model)
    graph.write(str(path), format="png")



from typing import Annotated, Any

import numpy as np
import numpy.typing as npt
from typing_extensions import Doc

from pdag._core import ParameterABC
from pdag._exec import ParameterId


def sample_parameter_values(
    input_parameters: dict[ParameterId, ParameterABC[Any]],
    n_samples: int,
    *,
    rng: np.random.Generator | None = None,
) -> list[dict[ParameterId, Any]]:
    unit_samples_cases = latin_hypercube_sampling(n_samples, len(input_parameters), rng=rng)
    return [
        {
            parameter_id: parameter.from_unit_interval(float(x))
            for (parameter_id, parameter), x in zip(input_parameters.items(), unit_samples, strict=True)
        }
        for unit_samples in unit_samples_cases
    ]


def latin_hypercube_sampling(
    n_samples: Annotated[int, Doc("Number of sample points.")],
    n_dimensions: Annotated[int, Doc("Number of dimensions (parameters).")],
    *,
    rng: Annotated[np.random.Generator | None, Doc("NumPy random number generator.")] = None,
) -> Annotated[npt.NDArray[np.float64], Doc("An array of shape (n_samples, n_dimensions) containing the samples.")]:
    """Generate Latin Hypercube Samples.

    Points are sampled from a unit hypercube in a way that ensures that each row and column contains exactly one sample.
    """
    # Initialize the sample matrix
    result = np.empty((n_samples, n_dimensions))
    # Divide the range [0, 1] into equal intervals
    cut = np.linspace(0, 1, n_samples + 1)

    rng = np.random.default_rng() if rng is None else rng

    # Fill the sample matrix
    for i in range(n_dimensions):
        # Randomly sample within each interval
        result[:, i] = rng.uniform(low=cut[:n_samples], high=cut[1 : n_samples + 1])
        # Shuffle to avoid any correlation between dimensions
        rng.shuffle(result[:, i])

    return result



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from pdag._core import CoreModel, ParameterABC, SubModelRelationship

if TYPE_CHECKING:
    from .model import StaticParameterId, TimeSeriesParameterId


def _search_for_parameter(
    model_path: tuple[str, ...],
    parameter_name: str,
    *,
    root_model: CoreModel,
) -> ParameterABC[Any]:
    if len(model_path) > 0:
        submodel_relationship = root_model.get_relationship(model_path[0])
        assert isinstance(submodel_relationship, SubModelRelationship)
        submodel = submodel_relationship.submodel
        return _search_for_parameter(
            model_path=model_path[1:],
            parameter_name=parameter_name,
            root_model=submodel,
        )

    return root_model.get_parameter(parameter_name)


def parameter_id_to_parameter(
    parameter_id: StaticParameterId | TimeSeriesParameterId,
    *,
    root_model: CoreModel,
) -> ParameterABC[Any]:
    return _search_for_parameter(
        model_path=parameter_id.model_path,
        parameter_name=parameter_id.name,
        root_model=root_model,
    )



import random
import string


def generate_random_string(length: int = 6) -> str:
    # Return random string
    return "".join(random.choices(string.ascii_letters + string.digits, k=length))  # noqa: S311



__all__ = [
    "Array",
    "ArrayRef",
    "BooleanParameter",
    "CategoricalParameter",
    "CollectionABC",
    "CollectionRef",
    "CoreModel",
    "ExecInfo",
    "FunctionRelationship",
    "Mapping",
    "MappingRef",
    "Module",
    "ParameterABC",
    "ParameterRef",
    "PydanticParameter",
    "RealParameter",
    "ReferenceABC",
    "RelationshipABC",
    "SubModelRelationship",
]

from .collection import Array, CollectionABC, Mapping
from .model import CoreModel, Module
from .parameter import BooleanParameter, CategoricalParameter, ParameterABC, PydanticParameter, RealParameter
from .reference import ArrayRef, CollectionRef, ExecInfo, MappingRef, ParameterRef, ReferenceABC
from .relationship import FunctionRelationship, RelationshipABC, SubModelRelationship



from abc import ABC, abstractmethod
from collections import defaultdict
from collections.abc import Hashable, Iterable
from dataclasses import dataclass, field
from enum import StrEnum
from typing import Annotated, Any, Self, cast

import numpy.typing as npt
from typing_extensions import Doc

from pdag._core import ExecInfo, FunctionRelationship
from pdag._core.model import CoreModel
from pdag._core.parameter import ParameterABC
from pdag._utils import topological_sort

from .utils import parameter_id_to_parameter

_model_path_doc = """\
Path to the model. The root model is represented by an empty tuple.

A sub-model below defined by relationship name `rel_name` is represented by a tuple ("rel_name",)`.
"""

type ModelPathType = Annotated[tuple[str, ...], Doc(_model_path_doc)]
type NodePathType = Annotated[tuple[str, ...], Doc("Path to the node.")]


@dataclass(frozen=True, slots=True)
class NodeIdMixin:
    model_path: ModelPathType
    name: str

    @property
    def node_path(self) -> NodePathType:
        return (*self.model_path, self.name)

    @property
    def node_path_str(self) -> str:
        return ".".join(self.node_path)

    @classmethod
    def from_node_path(cls, node_path: NodePathType) -> Self:
        return cls(model_path=node_path[:-1], name=node_path[-1])

    @classmethod
    def from_node_path_str(cls, node_path_str: str) -> Self:
        return cls.from_node_path(tuple(node_path_str.split(".")))


@dataclass(frozen=True, slots=True)
class ParameterIdMixin(NodeIdMixin):
    @property
    def parmaeter_path(self) -> NodePathType:
        return self.node_path

    @property
    def parameter_path_str(self) -> str:
        return self.node_path_str

    @classmethod
    def from_parameter_path(cls, parameter_path: NodePathType) -> Self:
        return cls.from_node_path(parameter_path)

    @classmethod
    def from_parameter_path_str(cls, parameter_path_str: str) -> Self:
        return cls.from_node_path_str(parameter_path_str)


@dataclass(frozen=True, slots=True)
class StaticParameterId(ParameterIdMixin):
    model_path: ModelPathType
    name: str


@dataclass(frozen=True, slots=True)
class TimeSeriesParameterId(ParameterIdMixin):
    model_path: ModelPathType
    name: str
    time_step: int


@dataclass(frozen=True, slots=True)
class StaticRelationshipId(NodeIdMixin):
    model_path: ModelPathType
    name: str


@dataclass(frozen=True, slots=True)
class TimeSeriesRelationshipId(NodeIdMixin):
    model_path: ModelPathType
    name: str
    time_step: int


class ExecInfoType(StrEnum):
    N_TIME_STEPS = "n_time_steps"
    TIME = "time"

    @classmethod
    def from_exec_info(cls, exec_info: ExecInfo) -> Self:
        return cls(exec_info.attribute)


# Type aliases
type ParameterId = StaticParameterId | TimeSeriesParameterId
type FunctionRelationshipInputId = ExecInfoType | ParameterId | dict[Hashable, ParameterId]
type RelationshipId = StaticRelationshipId | TimeSeriesRelationshipId
type NodeId = ParameterId | RelationshipId


@dataclass(slots=True)
class ConnectorABC(ABC):
    @abstractmethod
    def iter_parameter_ids(self) -> Iterable[ParameterId]:
        raise NotImplementedError


@dataclass(slots=True)
class ScalarConnector(ConnectorABC):
    parameter_id: ParameterId

    def iter_parameter_ids(self) -> Iterable[ParameterId]:
        yield self.parameter_id


@dataclass(slots=True)
class MappingConnector(ConnectorABC):
    parameter_ids: dict[Hashable, ParameterId]

    def iter_parameter_ids(self) -> Iterable[ParameterId]:
        yield from self.parameter_ids.values()


@dataclass(slots=True)
class MappingListConnector(ConnectorABC):
    parameter_ids: list[dict[Hashable, ParameterId]]

    def iter_parameter_ids(self) -> Iterable[ParameterId]:
        for mapping in self.parameter_ids:
            yield from mapping.values()


@dataclass(slots=True)
class ArrayConnector(ConnectorABC):
    parameter_ids: npt.NDArray[ParameterId]  # type: ignore[type-var]

    def iter_parameter_ids(self) -> Iterable[ParameterId]:
        yield from self.parameter_ids.flat


@dataclass(slots=True)
class FunctionRelationshipInfo:
    function_relationship: FunctionRelationship[Any, Any]
    input_parameter_info: dict[str, ConnectorABC | ExecInfoType]
    output_parameter_info: tuple[ConnectorABC, ...]


@dataclass(slots=True)
class ExecutionModel:
    parameter_ids: set[ParameterId]
    # SubModelRelationships should be flattened into FunctionRelationships
    relationship_infos: dict[RelationshipId, FunctionRelationshipInfo]
    input_parameter_id_to_relationship_ids: dict[ParameterId, set[RelationshipId]]
    relationship_id_to_output_parameter_ids: dict[RelationshipId, set[ParameterId]]
    # parent model output to sub-model input / sub-model output to parent model input
    port_mapping: dict[ParameterId, ParameterId]

    _core_model: CoreModel = field(repr=False, compare=False, kw_only=True)

    n_time_steps: int | None = None

    # Derived attributes
    relationship_id_to_input_parameter_ids: dict[RelationshipId, set[ParameterId]] = field(
        init=False,
        repr=False,
        compare=False,
    )
    output_parameter_id_to_relationship_ids: dict[ParameterId, set[RelationshipId]] = field(
        init=False,
        repr=False,
        compare=False,
    )
    port_mapping_inverse: dict[ParameterId, ParameterId] = field(init=False, repr=False, compare=False)

    _topologically_sorted_node_ids: list[NodeId] = field(init=False, repr=False, compare=False)

    def __post_init__(self) -> None:
        relationship_id_to_input_parameter_ids_dd: defaultdict[RelationshipId, set[ParameterId]] = defaultdict(
            set,
        )
        for input_parameter_id, relationship_ids in self.input_parameter_id_to_relationship_ids.items():
            for relationship_id in relationship_ids:
                relationship_id_to_input_parameter_ids_dd[relationship_id].add(input_parameter_id)
        self.relationship_id_to_input_parameter_ids = dict(relationship_id_to_input_parameter_ids_dd)

        output_parameter_id_to_relationship_ids_dd: defaultdict[ParameterId, set[RelationshipId]] = defaultdict(
            set,
        )
        for relationship_id, output_parameter_ids in self.relationship_id_to_output_parameter_ids.items():
            for output_parameter_id in output_parameter_ids:
                output_parameter_id_to_relationship_ids_dd[output_parameter_id].add(relationship_id)
        self.output_parameter_id_to_relationship_ids = dict(output_parameter_id_to_relationship_ids_dd)

        self.port_mapping_inverse = {value: key for key, value in self.port_mapping.items()}

        # Calculate topological sort
        dependencies_dd: defaultdict[NodeId, set[NodeId]] = defaultdict(
            set,
            cast("dict[NodeId, set[NodeId]]", self.input_parameter_id_to_relationship_ids)
            | cast("dict[NodeId, set[NodeId]]", self.relationship_id_to_output_parameter_ids),
        )
        for src, dest in self.port_mapping.items():
            dependencies_dd[src].add(dest)
        dependencies = dict(dependencies_dd)
        # Sort nodes topologically
        self._topologically_sorted_node_ids = topological_sort(dependencies)

    def input_parameter_ids(self) -> set[ParameterId]:
        return {
            parameter_id
            for parameter_id in self.parameter_ids
            if parameter_id not in self.output_parameter_id_to_relationship_ids
            and parameter_id not in self.port_mapping_inverse
        }

    def input_parameters(self) -> dict[ParameterId, ParameterABC[Any]]:
        return {
            parameter_id: parameter_id_to_parameter(parameter_id, root_model=self._core_model)
            for parameter_id in self.input_parameter_ids()
        }

    def parameters(self) -> dict[ParameterId, ParameterABC[Any]]:
        return {
            parameter_id: parameter_id_to_parameter(parameter_id, root_model=self._core_model)
            for parameter_id in self.parameter_ids
        }

    @property
    def topologically_sorted_node_ids(self) -> list[NodeId]:
        return self._topologically_sorted_node_ids



from collections import defaultdict
from collections.abc import Iterable, Mapping
from itertools import tee
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any

import polars as pl
import pyarrow as pa
from mpire import WorkerPool  # type: ignore[attr-defined]
from pyarrow import ipc
from rich.console import Console

import pdag
from pdag._exec.model import ExecutionModel, ParameterId

from .runner import _infinite_empty_dict_generator

console = Console()
err_console = Console(stderr=True)


def _result_to_df_rows(
    result: Mapping[pdag.StaticParameterId | pdag.TimeSeriesParameterId, Any],
    metadata: Mapping[str, Any] | None = None,
) -> list[dict[str, Any]]:
    # Containers for static and time-series parameters.
    static_data: dict[str, Any] = {}
    timeseries_data_dd: defaultdict[int, dict[str, Any]] = defaultdict(dict)
    for parameter_id, value in result.items():
        # We assume that static parameter_ids have only the `name` attribute,
        # while time series parameter_ids have `name` and `time_step` attributes.
        if isinstance(parameter_id, pdag.StaticParameterId):
            static_data[parameter_id.parameter_path_str] = value
        elif isinstance(parameter_id, pdag.TimeSeriesParameterId):
            ts = parameter_id.time_step
            timeseries_data_dd[ts][parameter_id.parameter_path_str] = value
    timeseries_data = dict(timeseries_data_dd)

    # For each time step, create a row that merges static and time-series parameters.
    metadata = dict(metadata) if metadata is not None else {}
    if timeseries_data:
        return [{"time_step": ts} | metadata | static_data | ts_params for ts, ts_params in timeseries_data.items()]
    # If there are no time-series parameters, just return the static data.
    return [metadata | static_data]


def _task(
    exec_model: pdag.ExecutionModel,
    case: Mapping[pdag.ParameterId, Any],
    metadata: Mapping[str, Any],
) -> list[dict[str, Any]]:
    result = pdag.execute_exec_model(exec_model, inputs=case)
    return _result_to_df_rows(result, metadata)


def _write_batch(batch: list[dict[str, Any]], writer: ipc.RecordBatchFileWriter, schema: pa.Schema) -> None:
    table = pa.Table.from_pylist(batch, schema=schema)
    writer.write_table(table)


def run_experiments(  # noqa: PLR0913
    exec_model: ExecutionModel,
    cases: Iterable[Mapping[ParameterId, Any]],
    *,
    metadata: Iterable[Mapping[str, Any]] | None = None,
    n_cases: int | None = None,
    delete_arrow_file: bool = True,
    parquet_file_path: str | Path,
) -> None:
    cases_warmup, cases = tee(cases)
    metadata_warmup, metadata = tee(metadata if metadata is not None else _infinite_empty_dict_generator())

    # Create a sample result to get the schema
    sample_result = _task(exec_model, case=next(iter(cases_warmup)), metadata=next(iter(metadata_warmup)))
    schema = pa.Table.from_pylist(sample_result).schema

    # Write the results to an Arrow file
    buffer = []
    batch_size = 1_000
    with TemporaryDirectory(delete=delete_arrow_file) as temp_dir:
        arrow_file_path = Path(temp_dir) / "results.arrow"

        with ipc.RecordBatchFileWriter(str(arrow_file_path), schema=schema) as writer:
            with WorkerPool(shared_objects=exec_model) as pool:
                console.log(f"Running experiments and writing to {arrow_file_path}...")
                for result in pool.imap_unordered(
                    _task,
                    ({"case": case, "metadata": meta} for case, meta in zip(cases, metadata, strict=False)),
                    iterable_len=n_cases,
                    progress_bar=True,
                ):
                    buffer.extend(result)
                    if len(buffer) >= batch_size:
                        _write_batch(buffer, writer, schema)
                        buffer.clear()

            if buffer:
                _write_batch(buffer, writer, schema)

            console.log(f"Finished running experiments. Arrow file written to {arrow_file_path}.")

        # Convert the generated Arrow file to Parquet
        # Experiences have shown that this is faster than writing directly to Parquet.
        with err_console.status("Converting Arrow to Parquet..."):
            lf = pl.scan_ipc(arrow_file_path)
            lf.sink_parquet(parquet_file_path)
        console.log(f"Converted arrow file to parquet file: {parquet_file_path}")

    if delete_arrow_file:
        console.log(f"Arrow file deleted: {arrow_file_path}")



"""Utility functions and classes for the pdag package."""

__all__ = [
    "InitArgsRecorder",
    "MultiDef",
    "MultiDefMeta",
    "MultiDefProtocol",
    "get_function_body",
    "merge_two_set_dicts",
    "multidef",
    "topological_sort",
]

from .ast_utils import get_function_body
from .dict_utils import merge_two_set_dicts
from .init_args_recorder import InitArgsRecorder
from .multidef import MultiDef, MultiDefMeta, MultiDefProtocol, multidef
from .topological_sort import topological_sort



from collections.abc import Generator, Iterable, Mapping
from typing import Any, Literal, overload

import polars as pl
from tqdm import tqdm

from pdag._exec import ExecutionModel, ParameterId, execute_exec_model

from .results import results_to_df


def _infinite_empty_dict_generator() -> Generator[dict[str, Any]]:
    while True:
        yield {}


@overload
def run_experiments(
    exec_model: ExecutionModel,
    cases: Iterable[Mapping[ParameterId, Any]],
    *,
    metadata: Iterable[Mapping[str, Any]],
    return_type: Literal["list"],
    n_cases: int | None = None,
) -> list[dict[ParameterId, Any]]: ...
@overload
def run_experiments(
    exec_model: ExecutionModel,
    cases: Iterable[Mapping[ParameterId, Any]],
    *,
    metadata: None = None,
    return_type: Literal["list"],
    n_cases: int | None = None,
) -> list[dict[ParameterId | str, Any]]: ...
@overload
def run_experiments(
    exec_model: ExecutionModel,
    cases: Iterable[Mapping[ParameterId, Any]],
    *,
    metadata: Iterable[Mapping[str, Any]] | None = None,
    return_type: Literal["polars"] = "polars",
    n_cases: int | None = None,
) -> pl.DataFrame: ...


def run_experiments(
    exec_model: ExecutionModel,
    cases: Iterable[Mapping[ParameterId, Any]],
    *,
    metadata: Iterable[Mapping[str, Any]] | None = None,
    return_type: Literal["list", "polars"] = "polars",
    n_cases: int | None = None,
) -> list[dict[ParameterId, Any]] | list[dict[ParameterId | str, Any]] | pl.DataFrame:
    results_iter = (
        execute_exec_model(
            exec_model,
            inputs=case,
        )
        | {"metadata": mtd}
        for case, mtd in tqdm(
            zip(cases, metadata if metadata is not None else _infinite_empty_dict_generator(), strict=False),
            total=n_cases,
        )
    )
    match return_type:
        case "list":
            return list(results_iter)
        case "polars":
            return results_to_df(results_iter)
        case _:
            msg = f"Invalid return_type: {return_type}"
            raise ValueError(msg)



from collections import defaultdict, deque
from collections.abc import Collection, Hashable, Mapping


def topological_sort[T: Hashable](dependencies: Mapping[T, Collection[T]]) -> list[T]:
    """Sort a graph of dependencies topologically.

    Given a dependency graph (a dict mapping a node to a collection of nodes that depend on it),
    perform a topological sort and return an ordered list of nodes.
    Raises an error if a cycle is detected.
    """
    indegree: defaultdict[T, int] = defaultdict(int)
    for node, deps in dependencies.items():
        indegree[node] = indegree.get(node, 0)
        for dep in deps:
            indegree[dep] += 1
    q = deque([node for node, deg in indegree.items() if deg == 0])
    order = []
    while q:
        node = q.popleft()
        order.append(node)
        for dep in dependencies.get(node, []):
            indegree[dep] -= 1
            if indegree[dep] == 0:
                q.append(dep)
    if len(order) != len(indegree):
        msg = "Cycle detected in relationship dependencies!"
        raise ValueError(msg)
    return order



from collections import deque
from collections.abc import Mapping
from typing import Any

import numpy as np

from .model import (
    ArrayConnector,
    ExecInfoType,
    ExecutionModel,
    MappingConnector,
    MappingListConnector,
    ParameterId,
    RelationshipId,
    ScalarConnector,
    StaticParameterId,
    StaticRelationshipId,
    TimeSeriesParameterId,
    TimeSeriesRelationshipId,
)


def _get_exec_info_value(
    exec_model: ExecutionModel,
    exec_info_type: ExecInfoType,
    relationship_id: RelationshipId,
) -> Any:
    match exec_info_type:
        case ExecInfoType.N_TIME_STEPS:
            return exec_model.n_time_steps
        case ExecInfoType.TIME:
            if isinstance(relationship_id, TimeSeriesRelationshipId):
                return relationship_id.time_step
            msg = f"Node {relationship_id} is not a TimeSeriesRelationshipId"
            raise ValueError(msg)


def execute_exec_model(  # noqa: C901, PLR0912, PLR0915
    exec_model: ExecutionModel,
    inputs: Mapping[ParameterId, Any],
) -> dict[ParameterId, Any]:
    # TODO: Exec only part of the model (in which case you need to sort the nodes again?)
    sorted_nodes_queue = deque(exec_model.topologically_sorted_node_ids)
    results: dict[ParameterId, Any] = {}
    while sorted_nodes_queue:
        node_id = sorted_nodes_queue.popleft()

        if node_id in results:
            continue

        if isinstance(node_id, StaticParameterId | TimeSeriesParameterId):
            if node_id in inputs:
                results[node_id] = inputs[node_id]
                continue
            if node_id in exec_model.port_mapping_inverse:
                parent_node_id = exec_model.port_mapping_inverse[node_id]
                if parent_node_id in inputs:
                    results[node_id] = inputs[parent_node_id]
                    continue
                if parent_node_id in results:
                    results[node_id] = results[parent_node_id]
                    continue
                msg = f"Parent node {parent_node_id} not in results"
                raise ValueError(msg)
            msg = f"Node {node_id} not in inputs or port_mapping"
            raise ValueError(msg)

        if isinstance(node_id, StaticRelationshipId | TimeSeriesRelationshipId):
            relationship_info = exec_model.relationship_infos[node_id]

            input_values: dict[str, Any] = {}
            for input_arg_name, connector_or_exec_info in relationship_info.input_parameter_info.items():
                if isinstance(connector_or_exec_info, ExecInfoType):
                    input_values[input_arg_name] = _get_exec_info_value(
                        exec_model,
                        connector_or_exec_info,
                        node_id,
                    )
                    continue
                if isinstance(connector_or_exec_info, ScalarConnector):
                    input_values[input_arg_name] = results[connector_or_exec_info.parameter_id]
                    continue
                if isinstance(connector_or_exec_info, MappingConnector):
                    input_values[input_arg_name] = {
                        key: results[param_id] for key, param_id in connector_or_exec_info.parameter_ids.items()
                    }
                    continue
                if isinstance(connector_or_exec_info, MappingListConnector):
                    input_values[input_arg_name] = [
                        {key: results[param_id] for key, param_id in mapping.items()}
                        for mapping in connector_or_exec_info.parameter_ids
                    ]
                    continue
                if isinstance(connector_or_exec_info, ArrayConnector):
                    value_array = np.empty(connector_or_exec_info.parameter_ids.shape, dtype=object)
                    for index, param_id in np.ndenumerate(connector_or_exec_info.parameter_ids):
                        value_array[index] = results[param_id]
                    input_values[input_arg_name] = value_array.tolist()
                    continue
                msg = f"Connector type {type(connector_or_exec_info)} is not supported."
                raise TypeError(msg)

            output_values = relationship_info.function_relationship(**input_values)

            if relationship_info.function_relationship.output_is_scalar:
                output_values = (output_values,)
            for connector, output_value in zip(
                relationship_info.output_parameter_info,
                output_values,
                strict=True,
            ):
                if isinstance(connector, ScalarConnector):
                    results[connector.parameter_id] = output_value
                    continue
                if isinstance(connector, MappingConnector):
                    for key, param_id in connector.parameter_ids.items():
                        results[param_id] = output_value[key]
                    continue
                if isinstance(connector, MappingListConnector):
                    for mapping, output_value_item in zip(connector.parameter_ids, output_value, strict=True):
                        for key, param_id in mapping.items():
                            results[param_id] = output_value_item[key]
                    continue
                if isinstance(connector, ArrayConnector):
                    for index, param_id in np.ndenumerate(connector.parameter_ids):
                        results[param_id] = np.asarray(output_value, dtype=object)[index]
                    continue
                msg = f"Connector type {type(connector)} is not supported."
                raise TypeError(msg)

    return results



from typing import Annotated, Literal

import pdag


class DiamondMdpModel(pdag.Model):
    """Diamond MDP model."""

    policy = pdag.CategoricalParameter("policy", categories=("left", "right"))
    location = pdag.CategoricalParameter("location", categories=("start", "left", "right", "end"), is_time_series=True)
    action = pdag.CategoricalParameter(
        "action",
        categories=("go_left", "go_right", "move_forward", "none"),
        is_time_series=True,
    )
    reward = pdag.RealParameter("reward", is_time_series=True)
    cumulative_reward = pdag.RealParameter("cumulative_reward")

    @pdag.relationship(at_each_time_step=True)
    @staticmethod
    def action_selection(  # noqa: PLR0911
        *,
        policy: Annotated[Literal["left", "right"], policy.ref()],
        location: Annotated[Literal["start", "left", "right", "end"], location.ref()],
    ) -> Annotated[Literal["go_left", "go_right", "move_forward", "none"], action.ref()]:
        match location, policy:
            case "start", "left":
                return "go_left"
            case "start", "right":
                return "go_right"
            case "left", "left":
                return "move_forward"
            case "left", "right":
                return "go_right"
            case "right", "right":
                return "move_forward"
            case "right", "left":
                return "go_left"
            case "end", _:
                return "none"

        msg = f"Invalid policy and location combination: {policy=}, {location=}"
        raise ValueError(msg)

    @pdag.relationship(at_each_time_step=True)
    @staticmethod
    def state_transition(  # noqa: C901, PLR0911
        *,
        location: Annotated[Literal["start", "left", "right", "end"], location.ref()],
        action: Annotated[Literal["go_left", "go_right", "move_forward", "none"], action.ref()],
    ) -> Annotated[Literal["start", "left", "right", "end"], location.ref(next=True)]:
        match location, action:
            case "start", "go_left":
                return "left"
            case "start", "go_right":
                return "right"
            case "start", "move_forward":
                return "start"
            case "left", "move_forward":
                return "end"
            case "left", "go_left":
                return "left"
            case "left", "go_right":
                return "right"
            case "right", "move_forward":
                return "end"
            case "right", "go_left":
                return "left"
            case "right", "go_right":
                return "right"
            case _, "none":
                return location
            case "end", _:
                return "end"

        msg = f"Invalid location and action combination: {location=}, {action=}"
        raise ValueError(msg)

    @pdag.relationship
    @staticmethod
    def initial_reward() -> Annotated[float, reward.ref(initial=True)]:
        return 0.0

    @pdag.relationship(at_each_time_step=True)
    @staticmethod
    def reward_function(
        *,
        previous_location: Annotated[
            Literal["start", "left", "right", "end"],
            location.ref(previous=True),
        ],
        action: Annotated[  # noqa: ARG004
            Literal["go_left", "go_right", "move_forward", "none"],
            action.ref(previous=True),
        ],
        location: Annotated[Literal["start", "left", "right", "end"], location.ref()],
    ) -> Annotated[float, reward.ref()]:
        if previous_location != "end" and location == "end":
            return 1.0
        return 0.0

    @pdag.relationship
    @staticmethod
    def cumulative_reward_calculation(
        *,
        reward: Annotated[list[float], reward.ref(all_time_steps=True)],
    ) -> Annotated[float, cumulative_reward.ref()]:
        return sum(reward)


if __name__ == "__main__":
    from rich import print  # noqa: A004

    core_model = DiamondMdpModel.to_core_model()
    print(core_model)
    exec_model = pdag.create_exec_model_from_core_model(core_model, n_time_steps=4)
    print(exec_model)
    results = pdag.execute_exec_model(
        exec_model,
        inputs={
            pdag.StaticParameterId((), "policy"): "left",
            pdag.TimeSeriesParameterId((), "location", 0): "start",
        },
    )

    print(results)



from abc import ABCMeta
from collections.abc import Callable, Hashable, Iterable
from typing import Any, Protocol

type IdentifierType = Hashable


class MultiDefStorage[K: IdentifierType, V](dict[K, V]):
    pass


# A custom dict that collects duplicate definitions in the class body.
class MultiDict[K, IT](dict[K, object]):
    def __setitem__(self, key: K, value: object) -> None:
        if isinstance(value, MultiDefStorage):
            msg = "Cannot assign a MultiDefStorage directly to a MultiDict."
            raise TypeError(msg)

        if hasattr(value, "__multidef_identifier__"):
            identifier: IT = value.__multidef_identifier__  # pyright: ignore[reportAttributeAccessIssue]
            if key in self:
                existing = super().__getitem__(key)
                # Only aggregate if the new value have our decorator marker.
                # We don't check for the existing value because it may already be a list.
                if isinstance(existing, MultiDefStorage):
                    assert all(hasattr(item, "__multidef_identifier__") for item in existing.values())
                    existing[identifier] = value
                else:
                    msg = f"Existing value for {key!r} is not a multidef object, cannot aggregate."
                    raise ValueError(msg)
            else:
                # First time we see this key, create a new MultiDefStorage
                super().__setitem__(key, MultiDefStorage({identifier: value}))

        else:
            super().__setitem__(key, value)


def _to_unique_name(attr: str, identifier_str: str) -> str:
    return f"__multidef_{attr}_{identifier_str}__"


def _to_identifier_str(identifier: IdentifierType) -> str:
    """Try to convert the identifier to a string that can be used as an attribute name."""
    if isinstance(identifier, str):
        return identifier
    if isinstance(identifier, Iterable):
        return "_".join(map(_to_identifier_str, identifier))

    # Fallback to the default string representation.
    # If this cannot be used as an attribute name, the user will get an error.
    return str(identifier)


class MultiDefProtocol[K: IdentifierType, T](Protocol):
    def __getitem__(self, identifier: K) -> T: ...


# The decorator simply attaches an identifier to the function.
# The type hints are not correct here, but will be corrected by the metaclass.
def multidef[K: IdentifierType, T](
    identifier: K,
) -> Callable[[T], MultiDefProtocol[K, T]]:
    _unique_name = _to_unique_name("_", _to_identifier_str(identifier))
    if not _unique_name.isidentifier():
        msg = f"Invalid identifier: {identifier!r}. Attribute name could be {_unique_name!r}."
        raise ValueError(msg)

    def decorator[T2](
        obj: T2,
    ) -> MultiDefProtocol[IdentifierType, T2]:
        try:
            obj.__multidef_identifier__ = identifier  # type: ignore[attr-defined]
        except AttributeError as e:
            msg = "multidef can only be used on objects that support attribute assignment."
            raise TypeError(msg) from e

        return obj  # type: ignore[return-value]

    return decorator


class MultiDefMeta(ABCMeta):
    @classmethod
    def __prepare__(
        metacls,  # noqa: N804
        name: str,
        bases: tuple[type[Any], ...],
        /,
        **kwargs: Any,
    ) -> MultiDict[str, IdentifierType]:
        return MultiDict()

    def __new__(
        metacls,
        name: str,
        bases: tuple[type[Any], ...],
        namespace: dict[str, Any],
    ) -> type:
        # Create the class first using a normal dict
        cls = super().__new__(metacls, name, bases, dict(namespace))
        # Now process duplicate definitions in the original namespace.
        for attr, value in namespace.items():
            if isinstance(value, MultiDefStorage):
                identifier_to_unique_name = {
                    identifier: _to_unique_name(attr, _to_identifier_str(identifier)) for identifier in value
                }
                for identifier, obj in value.items():
                    # Only process objects that have our marker
                    unique_name = identifier_to_unique_name[identifier]
                    setattr(cls, unique_name, obj)

        return cls


# Example class using the metaclass and the decorator.
class MultiDef(metaclass=MultiDefMeta):
    def get_name(self, attr: str, identifier: IdentifierType) -> str:
        return _to_unique_name(attr, _to_identifier_str(identifier))

    def get_def(self, attr: str, identifier: IdentifierType) -> Any:
        return getattr(self, self.get_name(attr, identifier))

    def get_defs(self, attr: str) -> dict[IdentifierType, Any]:
        return {identifier: self.get_def(attr, identifier) for identifier in super().__getattribute__(attr)}

    def __getattribute__(self, name: str) -> Any:
        value = super().__getattribute__(name)
        if isinstance(value, MultiDefStorage):
            return self.get_defs(name)
        return value



from collections.abc import Hashable
from collections.abc import Mapping as MappingABC
from types import EllipsisType
from typing import Any

import numpy as np
import numpy.typing as npt

from pdag._core import ReferenceABC
from pdag._core.collection import Array, CollectionABC, Mapping
from pdag._core.model import CoreModel
from pdag._core.parameter import ParameterABC
from pdag._core.reference import CollectionRef, ParameterRef

from .model import (
    ArrayConnector,
    ConnectorABC,
    MappingConnector,
    MappingListConnector,
    ModelPathType,
    ParameterId,
    ScalarConnector,
    StaticParameterId,
    TimeSeriesParameterId,
)


def _mapping_to_array[T](
    mapping: MappingABC[tuple[int, ...], T],
    shape: tuple[int, ...],
    *,
    error_on_missing: bool = False,
) -> npt.NDArray[T]:  # type: ignore[type-var]
    array = np.empty(shape, dtype=object)
    for key, value in mapping.items():
        array[key] = value
    if error_on_missing and np.any(array == None):  # noqa: E711
        msg = "Missing values in mapping."
        raise ValueError(msg)
    return array


def resolve_ref(  # noqa: PLR0913
    ref: ReferenceABC,
    *,
    core_model: CoreModel,
    model_path: ModelPathType,
    time_series_relationship: bool,
    n_time_steps: int | None = None,
    time_step: int | None = None,
) -> ConnectorABC:
    obj = core_model.get_object_from_ref(ref)
    if isinstance(obj, ParameterABC):
        assert isinstance(ref, ParameterRef)
        if time_series_relationship:
            assert time_step is not None
            return _resolve_parameter_ref_in_time_series_relationship(
                ref=ref,
                model_path=model_path,
                parameter=obj,
                time_step=time_step,
            )
        assert n_time_steps is not None
        return _resolve_parameter_ref_in_static_relationship(
            ref=ref,
            model_path=model_path,
            parameter=obj,
            n_time_steps=n_time_steps,
        )
    if isinstance(obj, CollectionABC):
        assert isinstance(ref, CollectionRef)
        if time_series_relationship:
            assert time_step is not None
            return _resolve_collection_ref_in_time_series_relationship(
                ref=ref,
                model_path=model_path,
                collection=obj,
                time_step=time_step,
            )
        assert n_time_steps is not None
        return _resolve_collection_ref_in_static_relationship(
            ref=ref,
            model_path=model_path,
            collection=obj,
            n_time_steps=n_time_steps,
        )

    raise NotImplementedError


def _resolve_parameter_ref_in_time_series_relationship(
    ref: ParameterRef,
    *,
    model_path: ModelPathType,
    parameter: ParameterABC[Any],
    time_step: int,
) -> ConnectorABC:
    assert isinstance(parameter.name, str)
    if ref.normal:
        param_time_step = time_step
    elif ref.previous:
        param_time_step = time_step - 1
    elif ref.next:
        param_time_step = time_step + 1
    else:
        msg = "Unsupported reference type."
        raise ValueError(msg)

    if parameter.is_time_series:
        input_parameter_id: ParameterId = TimeSeriesParameterId(
            model_path=model_path,
            name=parameter.name,
            time_step=param_time_step,
        )
    else:
        input_parameter_id = StaticParameterId(
            model_path=model_path,
            name=parameter.name,
        )

    return ScalarConnector(parameter_id=input_parameter_id)


def _resolve_collection_ref_in_time_series_relationship(
    ref: CollectionRef[Hashable],
    *,
    model_path: ModelPathType,
    collection: CollectionABC[Any, ParameterABC[Any]],
    time_step: int,
) -> ConnectorABC:
    if collection.is_time_series():
        if ref.normal:
            param_time_step = time_step
        elif ref.previous:
            param_time_step = time_step - 1
        elif ref.next:
            param_time_step = time_step + 1
        else:
            msg = "Unsupported reference type."
            raise ValueError(msg)

        parameter_ids: dict[Hashable, ParameterId] = {
            key: TimeSeriesParameterId(
                model_path=model_path,
                name=parameter.name,
                time_step=param_time_step,
            )
            for key, parameter in collection.items()
        }
    else:
        parameter_ids = {
            key: StaticParameterId(model_path=model_path, name=parameter.name) for key, parameter in collection.items()
        }

    if isinstance(collection, Mapping):
        return _filter(MappingConnector(parameter_ids=parameter_ids), key=ref.key)
    if isinstance(collection, Array):
        return ArrayConnector(
            parameter_ids=_mapping_to_array(
                parameter_ids,  # type: ignore[arg-type]
                shape=collection.shape,
                error_on_missing=True,
            ),
        )

    msg = f"Unsupported collection type: {collection}"
    raise ValueError(msg)


def _resolve_parameter_ref_in_static_relationship(
    ref: ParameterRef,
    *,
    model_path: ModelPathType,
    parameter: ParameterABC[Any],
    n_time_steps: int,
) -> ConnectorABC:
    assert isinstance(parameter.name, str)

    if parameter.is_time_series:
        if ref.all_time_steps:
            parameter_ids: list[ParameterId] = [
                TimeSeriesParameterId(
                    model_path=model_path,
                    name=parameter.name,
                    time_step=time_step,
                )
                for time_step in range(n_time_steps)
            ]
            return ArrayConnector(parameter_ids=np.array(parameter_ids))
        if ref.initial:
            return ScalarConnector(
                parameter_id=TimeSeriesParameterId(
                    model_path=model_path,
                    name=parameter.name,
                    time_step=0,
                ),
            )
        msg = "Time-series parameter reference in static relationship must be either 'all_time_steps' or 'initial'."
        raise ValueError(msg)

    return ScalarConnector(
        parameter_id=StaticParameterId(model_path=model_path, name=parameter.name),
    )


def _resolve_collection_ref_in_static_relationship(
    ref: CollectionRef[Hashable],
    *,
    model_path: ModelPathType,
    collection: CollectionABC[Hashable, ParameterABC[Any]],
    n_time_steps: int,
) -> ConnectorABC:
    if collection.is_time_series():
        if ref.all_time_steps:
            parameter_ids_mlc: list[dict[Hashable, ParameterId]] = [
                {
                    key: TimeSeriesParameterId(
                        model_path=model_path,
                        name=parameter.name,
                        time_step=time_step,
                    )
                    for key, parameter in collection.items()
                }
                for time_step in range(n_time_steps)
            ]
            connector: ConnectorABC = MappingListConnector(
                parameter_ids=parameter_ids_mlc,
            )
        elif ref.initial:
            parameter_ids: dict[Hashable, ParameterId] = {
                key: TimeSeriesParameterId(
                    model_path=model_path,
                    name=parameter.name,
                    time_step=0,
                )
                for key, parameter in collection.items()
            }
            connector = MappingConnector(parameter_ids=parameter_ids)
        else:
            msg = (
                "Time-series collection reference in static relationship must be either 'all_time_steps' or 'initial'."
            )
            raise ValueError(msg)
    else:
        parameter_ids = {
            key: StaticParameterId(model_path=model_path, name=parameter.name) for key, parameter in collection.items()
        }
        connector = MappingConnector(parameter_ids=parameter_ids)

    if isinstance(collection, Mapping):
        return _filter(connector, key=ref.key)
    if isinstance(collection, Array):
        if isinstance(connector, MappingListConnector):
            # Time-series of an array collection
            return ArrayConnector(
                parameter_ids=np.array(
                    [
                        _mapping_to_array(
                            mapping,  # type: ignore[arg-type]
                            shape=collection.shape,
                            error_on_missing=True,
                        )
                        for mapping in connector.parameter_ids
                    ],
                ),
            )
        return ArrayConnector(
            parameter_ids=_mapping_to_array(
                parameter_ids,  # type: ignore[arg-type]
                shape=collection.shape,
                error_on_missing=True,
            ),
        )

    msg = f"Unsupported collection type: {collection}"
    raise ValueError


def _filter(connector: ConnectorABC, key: Hashable | None = None) -> ConnectorABC:
    if key is None:
        return connector
    if isinstance(connector, MappingConnector):
        if isinstance(key, str) or (isinstance(key, tuple) and all(isinstance(k, str) for k in key)):
            return ScalarConnector(parameter_id=connector.parameter_ids[key])
        if isinstance(key, tuple):
            return MappingConnector(
                parameter_ids=_filter_mapping(connector.parameter_ids, key),  # type: ignore[arg-type]
            )
        msg = f"Unsupported key type: {key}"
        raise ValueError(msg)
    if isinstance(connector, ArrayConnector):
        msg = "Filtering an array connector is not supported."
        raise NotImplementedError(msg)
    if isinstance(connector, MappingListConnector):
        sample_key = next(iter(connector.parameter_ids[0]))
        if isinstance(sample_key, str):
            return ArrayConnector(
                parameter_ids=np.array(
                    [mapping[key] for mapping in connector.parameter_ids],
                ),
            )
        if isinstance(sample_key, tuple):
            return MappingListConnector(
                parameter_ids=_filter_mapping(connector.parameter_ids, key),  # type: ignore[arg-type]
            )
        msg = f"Unsupported key type: {sample_key}"
        raise ValueError(msg)
    msg = f"Unsupported connector type: {connector}"
    raise ValueError(msg)


def _filter_mapping[T](
    mapping: MappingABC[tuple[str, ...], T],
    key: tuple[str | EllipsisType, ...],
) -> dict[str | tuple[str, ...], T]:
    mapping_filtered = {
        key_in_mapping: value
        for key_in_mapping, value in mapping.items()
        if all(
            (k_in_mapping == k_provided) or (k_provided is Ellipsis)
            for k_in_mapping, k_provided in zip(key_in_mapping, key, strict=True)
        )
    }

    # Reduce the dimension of the key if the value is provided
    def _reduce_key(key_in_mapping: tuple[str, ...]) -> str | tuple[str, ...]:
        reduced_key = tuple(k for k, k_provided in zip(key_in_mapping, key, strict=True) if k_provided is Ellipsis)
        if len(reduced_key) == 1:
            return reduced_key[0]
        return reduced_key

    return {_reduce_key(key_in_mapping): value for key_in_mapping, value in mapping_filtered.items()}



from typing import Annotated

from pydantic import BaseModel

import pdag

COLS = 3
ROWS = 3


class Cell(BaseModel):
    """A cell in a grid."""

    col: Annotated[int, pdag.RealParameter("col", lower_bound=0, upper_bound=COLS - 1)]
    row: Annotated[int, pdag.RealParameter("row", lower_bound=0, upper_bound=ROWS - 1)]


INITIAL_LOCATION = Cell(col=0, row=0)
EXIT = Cell(col=COLS - 1, row=ROWS - 1)


class AgentPolicy(BaseModel):
    """Agent policy for the treasure model."""

    type: Annotated[str, pdag.CategoricalParameter("type", categories=("sweep", "shortest"))]


class TreasureModel(pdag.Model):
    # Unknown parameters
    treasure_location = pdag.PydanticParameter("treasure_location", Cell)

    # Known parameters
    time_limit = pdag.RealParameter("time_limit")

    agent_location = pdag.PydanticParameter("agent_location", Cell, is_time_series=True)

    agent_policy = pdag.PydanticParameter("agent_policy", AgentPolicy)
    agent_action = pdag.CategoricalParameter(
        "agent_action",
        categories=("move_up", "move_down", "move_left", "move_right", "exit"),
        is_time_series=True,
    )

    treasure_found = pdag.BooleanParameter("treasure_found", is_time_series=True)
    exited = pdag.BooleanParameter("exited", is_time_series=True)
    time_limit_reached = pdag.BooleanParameter("time_limit_reached", is_time_series=True)

    @pdag.relationship(at_each_time_step=True)
    @staticmethod
    def determine_action(  # noqa: C901, PLR0911
        agent_location: Annotated[Cell, agent_location.ref()],
        agent_policy: Annotated[AgentPolicy, agent_policy.ref()],
    ) -> Annotated[str, agent_action.ref()]:
        """Determine the agent's action based on the policy."""
        match agent_policy.type:
            case "sweep":
                # Simple sweeping strategy
                match agent_location.row % 2:
                    case 0:  # Even row
                        if agent_location.col < COLS - 1:
                            return "move_right"
                        if agent_location.row < ROWS - 1:
                            return "move_down"
                        return "exit"
                    case 1:  # Odd row
                        if agent_location.col > 0:
                            return "move_left"
                        if agent_location.row < ROWS - 1:
                            return "move_down"
                        msg = f"Invalid location for sweeping strategy: {agent_location}"
                        raise ValueError(msg)
                    case _:
                        msg = f"Invalid location for sweeping strategy: {agent_location}"
                        raise ValueError(msg)
            case "shortest":
                # Shortest path strategy (placeholder logic)
                if agent_location.col < EXIT.col:
                    return "move_right"
                if agent_location.row < EXIT.row:
                    return "move_down"
                return "exit"
            case _:
                msg = f"Unknown agent policy type: {agent_policy.type}"
                raise ValueError(msg)

    @pdag.relationship()
    @staticmethod
    def initial_agent_location() -> Annotated[Cell, agent_location.ref(initial=True)]:
        """Initial agent location."""  # noqa: D401
        return INITIAL_LOCATION

    @pdag.relationship(at_each_time_step=True)
    @staticmethod
    def update_agent_location(  # noqa: C901
        agent_location: Annotated[Cell, agent_location.ref()],
        agent_action: Annotated[str, agent_action.ref()],
        exited: Annotated[bool, exited.ref()],
        time_limit_reached: Annotated[bool, time_limit_reached.ref()],
    ) -> Annotated[Cell, agent_location.ref(next=True)]:
        """Update the agent's location based on the action."""
        if exited or time_limit_reached:
            return agent_location
        new_location = Cell(col=agent_location.col, row=agent_location.row)
        match agent_action:
            case "move_up":
                if new_location.row > 0:
                    new_location.row -= 1
            case "move_down":
                if new_location.row < ROWS - 1:
                    new_location.row += 1
            case "move_left":
                if new_location.col > 0:
                    new_location.col -= 1
            case "move_right":
                if new_location.col < COLS - 1:
                    new_location.col += 1
            case "exit":
                pass
            case _:
                msg = f"Unknown agent action: {agent_action}"
                raise ValueError(msg)
        return new_location

    @pdag.relationship()
    @staticmethod
    def initial_exited() -> Annotated[bool, exited.ref(initial=True)]:
        """Initial exited state."""  # noqa: D401
        return False

    @pdag.relationship(at_each_time_step=True)
    @staticmethod
    def update_exited(
        agent_location: Annotated[Cell, agent_location.ref()],
        agent_action: Annotated[str, agent_action.ref()],
        time_limit_reached: Annotated[bool, time_limit_reached.ref()],
    ) -> Annotated[bool, exited.ref(next=True)]:
        """Check if the agent has exited."""
        return agent_location == EXIT and not time_limit_reached and agent_action == "exit"

    @pdag.relationship()
    @staticmethod
    def initial_treasure_found() -> Annotated[bool, treasure_found.ref(initial=True)]:
        """Initial treasure found state."""  # noqa: D401
        return False

    @pdag.relationship(at_each_time_step=True)
    @staticmethod
    def check_treasure_found(
        agent_location: Annotated[Cell, agent_location.ref()],
        treasure_location: Annotated[Cell, treasure_location.ref()],
        time_limit_reached: Annotated[bool, time_limit_reached.ref()],
        treasure_found: Annotated[bool, treasure_found.ref(previous=True)],
    ) -> Annotated[bool, treasure_found.ref()]:
        """Check if the agent has found the treasure."""
        return treasure_found or (agent_location == treasure_location and not time_limit_reached)

    @pdag.relationship(at_each_time_step=True)
    @staticmethod
    def check_time_limit_reached(
        time_limit: Annotated[float, time_limit.ref()],
        time: Annotated[int, pdag.ExecInfo("time")],
    ) -> Annotated[bool, time_limit_reached.ref()]:
        """Check if the time limit has been reached."""
        return time >= time_limit


if __name__ == "__main__":
    from rich import print  # noqa: A004

    core_model = TreasureModel.to_core_model()
    print(core_model)
    exec_model = pdag.create_exec_model_from_core_model(core_model, n_time_steps=10)
    print(exec_model)
    print(list(exec_model.input_parameter_ids()))
    results = pdag.execute_exec_model(
        exec_model,
        inputs={
            pdag.StaticParameterId((), "treasure_location"): Cell(col=1, row=1),
            pdag.StaticParameterId((), "time_limit"): 10.0,
            pdag.StaticParameterId((), "agent_policy"): AgentPolicy(type="sweep"),
        },
    )
    print(results)



__all__ = [
    "distance_constrained_sampling",
    "results_to_df",
    "run_experiments",
    "sample_parameter_values",
]
from .cases import sample_parameter_values
from .distance_sampling import distance_constrained_sampling
from .results import results_to_df
from .runner import run_experiments



from abc import ABC, abstractmethod
from collections.abc import Callable, Iterable
from dataclasses import dataclass, field
from types import EllipsisType
from typing import TYPE_CHECKING, ClassVar

from pdag._utils import InitArgsRecorder

from .reference import ExecInfo, ReferenceABC

if TYPE_CHECKING:
    from .model import CoreModel


@dataclass
class RelationshipABC(ABC, InitArgsRecorder):
    type: ClassVar[str] = "relationship"
    _name: str | EllipsisType
    at_each_time_step: bool = field(default=False, kw_only=True)

    def name_is_set(self) -> bool:
        return isinstance(self._name, str)

    @property
    def name(self) -> str:
        if isinstance(self._name, EllipsisType):
            msg = "Relationship name has not been set."
            raise ValueError(msg)  # noqa: TRY004
        return self._name

    @name.setter
    def name(self, value: str) -> None:
        self._name = value

    @abstractmethod
    def is_hydrated(self) -> bool:
        raise NotImplementedError

    @abstractmethod
    def iter_input_refs(self) -> Iterable[ReferenceABC | ExecInfo]:
        raise NotImplementedError

    @abstractmethod
    def iter_output_refs(self) -> Iterable[ReferenceABC]:
        raise NotImplementedError

    @property
    def includes_past(self) -> bool:
        return any(param_ref.previous for param_ref in self.iter_input_refs() if isinstance(param_ref, ReferenceABC))

    @property
    def includes_future(self) -> bool:
        return any(param_ref.next for param_ref in self.iter_output_refs())


@dataclass
class FunctionRelationship[**P, T](RelationshipABC):
    type: ClassVar[str] = "function"
    inputs: dict[str, ReferenceABC | ExecInfo] = field(kw_only=True)
    outputs: list[ReferenceABC] = field(kw_only=True)
    function_body: str = field(kw_only=True)
    output_is_scalar: bool = field(kw_only=True)
    _function: Callable[P, T] | None = field(default=None, compare=False, kw_only=True)

    def is_hydrated(self) -> bool:
        return self._function is not None

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> T:
        if self._function is None:
            msg = f"Function relationship {self.name} has not been hydrated."
            raise ValueError(msg)
        return self._function(*args, **kwargs)

    def iter_input_refs(self) -> Iterable[ReferenceABC | ExecInfo]:
        return self.inputs.values()

    def iter_output_refs(self) -> Iterable[ReferenceABC]:
        return self.outputs


@dataclass
class SubModelRelationship(RelationshipABC):
    type: ClassVar[str] = "submodel"
    submodel_name: str
    inputs: dict[
        ReferenceABC,
        ReferenceABC,
    ]  # sub-model parameter ref -> parent model parameter ref
    outputs: dict[
        ReferenceABC,
        ReferenceABC,
    ]  # sub-model parameter ref -> parent model parameter ref
    _submodel: "CoreModel | None" = field(default=None, compare=False, kw_only=True)

    def is_hydrated(self) -> bool:
        return self._submodel is not None

    def iter_input_refs(self) -> Iterable[ReferenceABC]:
        return self.inputs.values()

    def iter_output_refs(self) -> Iterable[ReferenceABC]:
        return self.outputs.values()

    @property
    def submodel(self) -> "CoreModel":
        if self._submodel is None:
            msg = f"Sub-model relationship {self.name} has not been hydrated."
            raise ValueError(msg)
        return self._submodel



from collections import defaultdict
from collections.abc import Callable, Iterable, Mapping
from typing import Any

import polars as pl

import pdag
from pdag._utils.random_string import generate_random_string


def results_to_df(
    results: Iterable[Mapping[pdag.StaticParameterId | pdag.TimeSeriesParameterId | str, Any]],
    case_id_factory: Callable[[], str] = generate_random_string,
) -> pl.DataFrame:
    rows: list[dict[str, Any]] = []
    for res in results:
        case_id = case_id_factory()
        # Containers for static and time-series parameters.
        static_data: dict[str, Any] = {}
        timeseries_data_dd: defaultdict[int, dict[str, Any]] = defaultdict(dict)
        for parameter_id, value in res.items():
            # We assume that static parameter_ids have only the `name` attribute,
            # while time series parameter_ids have `name` and `time_step` attributes.
            if isinstance(parameter_id, pdag.StaticParameterId):
                static_data[parameter_id.parameter_path_str] = value
            elif isinstance(parameter_id, pdag.TimeSeriesParameterId):
                ts = parameter_id.time_step
                timeseries_data_dd[ts][parameter_id.parameter_path_str] = value

        timeseries_data = dict(timeseries_data_dd)

        # For each time step, create a row that merges static and time-series parameters.
        metadata: dict[str, Any] = res.get("metadata", {})
        for ts, ts_params in sorted(timeseries_data.items()):
            row = {"case_id": case_id, "time_step": ts} | metadata
            # Include static parameters (repeated for each time step)
            row.update(static_data)
            # Include time-series parameters for the current time step
            row.update(ts_params)
            rows.append(row)

    return pl.DataFrame(rows)
